@@ -38,8 +38,15 @@
 #include "config.h"
 #endif
 
+#ifdef _MSC_VER
+/* MSVC complains about sprintf being deprecated in favor of sprintf_s */
+# define _CRT_SECURE_NO_WARNINGS
+/* MSVC complains about strdup being deprecated in favor of _strdup */
+# define _CRT_NONSTDC_NO_DEPRECATE
+#endif
+
 #include "amqp_private.h"
-#include "amqp_timer.h"
+#include "amqp_time.h"
 #include <assert.h>
 #include <stdarg.h>
 #include <stdint.h>
@@ -75,7 +82,8 @@ static const char *base_error_strings[] = {
   "heartbeat timeout, connection closed",/* AMQP_STATUS_HEARTBEAT_TIMEOUT       -0x000F */
   "unexpected protocol state",          /* AMQP_STATUS_UNEXPECTED STATE         -0x0010 */
   "socket is closed",                   /* AMQP_STATUS_SOCKET_CLOSED            -0x0011 */
-  "socket already open"                 /* AMQP_STATUS_SOCKET_INUSE             -0x0012 */
+  "socket already open",                /* AMQP_STATUS_SOCKET_INUSE             -0x0012 */
+  "unsupported sasl method requested"   /* AMQP_STATUS_BROKER_UNSUPPORTED_SASL_METHOD -0x0013 */
 };
 
 static const char *tcp_error_strings[] = {
@@ -190,23 +198,22 @@ int amqp_basic_publish(amqp_connection_state_t state,
   m.immediate = immediate;
   m.ticket = 0;
 
-  if (amqp_heartbeat_enabled(state)) {
-    uint64_t current_timestamp = amqp_get_monotonic_timestamp();
-    if (0 == current_timestamp) {
-      return AMQP_STATUS_TIMER_FAILURE;
-    }
-
-    if (current_timestamp > state->next_recv_heartbeat) {
-      res = amqp_try_recv(state, current_timestamp);
-      if (AMQP_STATUS_TIMEOUT == res) {
-        return AMQP_STATUS_HEARTBEAT_TIMEOUT;
-      } else if (AMQP_STATUS_OK != res) {
-        return res;
-      }
+  /* TODO(alanxz): this heartbeat check is happening in the wrong place, it
+   * should really be done in amqp_try_send/writev */
+  res = amqp_time_has_past(state->next_recv_heartbeat);
+  if (AMQP_STATUS_TIMER_FAILURE == res) {
+    return res;
+  } else if (AMQP_STATUS_TIMEOUT == res) {
+    res = amqp_try_recv(state);
+    if (AMQP_STATUS_TIMEOUT == res) {
+      return AMQP_STATUS_HEARTBEAT_TIMEOUT;
+    } else if (AMQP_STATUS_OK != res) {
+      return res;
     }
   }
 
-  res = amqp_send_method(state, channel, AMQP_BASIC_PUBLISH_METHOD, &m);
+  res = amqp_send_method_inner(state, channel, AMQP_BASIC_PUBLISH_METHOD, &m,
+                               AMQP_SF_MORE);
   if (res < 0) {
     return res;
   }
@@ -222,7 +229,7 @@ int amqp_basic_publish(amqp_connection_state_t state,
   f.payload.properties.body_size = body.len;
   f.payload.properties.decoded = (void *) properties;
 
-  res = amqp_send_frame(state, &f);
+  res = amqp_send_frame_inner(state, &f, AMQP_SF_MORE);
   if (res < 0) {
     return res;
   }
@@ -230,6 +237,7 @@ int amqp_basic_publish(amqp_connection_state_t state,
   body_offset = 0;
   while (body_offset < body.len) {
     size_t remaining = body.len - body_offset;
+    int flagz;
 
     if (remaining == 0) {
       break;
@@ -240,12 +248,14 @@ int amqp_basic_publish(amqp_connection_state_t state,
     f.payload.body_fragment.bytes = amqp_offset(body.bytes, body_offset);
     if (remaining >= usable_body_payload_size) {
       f.payload.body_fragment.len = usable_body_payload_size;
+      flagz = AMQP_SF_MORE;
     } else {
       f.payload.body_fragment.len = remaining;
+      flagz = AMQP_SF_NONE;
     }
 
     body_offset += f.payload.body_fragment.len;
-    res = amqp_send_frame(state, &f);
+    res = amqp_send_frame_inner(state, &f, flagz);
     if (res < 0) {
       return res;
     }
